import asyncio
from loguru import logger
from random import random
import numpy as np


async def solve(workers: asyncio.Queue[int], input: int) -> int:
    worker = await workers.get()
    logger.info(f"Task start, {input=} assigned to {worker=}.")
    progress = 0

    while progress < 1:
        work = random()
        await asyncio.sleep(work)
        progress += work
        logger.info(f"{progress=:.2f}")

    logger.info(f"Task finished, final {progress=:.2f}, {worker=} queued.")
    await workers.put(worker)
    return input * 2


async def orchestrator(n_workers: int, inputs: list[int]):
    workers = asyncio.Queue()
    for i in range(1, n_workers + 1):
        await workers.put(i)
    logger.info("Worker queue created.")

    async with asyncio.TaskGroup() as tg:
        tasks = [tg.create_task(solve(workers, i)) for i in inputs]

    results = np.vstack([task.result() for task in tasks])

    logger.info(f"All tasks finished, {results=}, {results.shape=}")


def main():
    inputs = [i for i in range(100, 105)]
    asyncio.run(orchestrator(3, inputs), debug=True)


if __name__ == "__main__":
    main()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from datetime import date
from typing import Any

# from sphinx import pygments_styles
from antcal import __version__

project = "AntCal"
copyright = f"{date.today().year}, Atlanswer"
author = "Atlanswer"
version = __version__
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "autodoc2",
    "sphinx_rtd_theme",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "pyaedt": ("https://aedt.docs.pyansys.com/version/stable/", None),
}

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_favicon = "_static/favicon.ico"
html_theme_options: dict[str, Any] = {
    "collapse_navigation": False,
    "navigation_depth": 4,
}
pygments_styles = "one-dark"
# pygments_dark_styles = "github-dark"

# -- myst --------------------------------------------------------------------
myst_enable_extensions = ["colon_fence", "dollarmath", "fieldlist"]
myst_heading_anchors = 2
myst_dmath_allow_space = False
myst_dmath_allow_digits = False

# -- autodoc2 ----------------------------------------------------------------
autodoc2_packages = ["../src/antcal"]
autodoc2_render_plugin = "myst"
# autodoc2_docstring_parser_regexes = [(r".*", "myst"),]



"""Slotted Patch.

Reference: A. Papathanasopoulos,
P. A. Apostolopoulos and Y. Rahmat-Samii,
"Optimization Assisted by Neural Network-Based
Machine Learning in Electromagnetic Applications,"
_IEEE Transactions on Antennas and Propagation_,
Jan. 2023, doi: 10.1109/TAP.2023.3269883.
"""

# %%
import asyncio
import time

import numpy as np
import numpy.typing as npt
from loguru import logger
from pyaedt.hfss import Hfss
from pyaedt.modeler.cad.elements3d import FacePrimitive
from pyaedt.modeler.cad.object3d import Object3d
from pyaedt.modeler.modeler3d import Modeler3D
from pyaedt.modules.solutions import SolutionData
from pyaedt.modules.SolveSetup import SetupHFSS

from antcal.pyaedt.hfss import (
    check_materials,
    new_hfss_session,
    update_variables,
)

# %%
N_DIMS_SLOTTED_PATCH = 10
VAR_BOUNDS = (
    [30.0, 30.0, 1.0, 1.0, 1.0, -10.0, 7.5, 5.0, 0.0, 0.0],
    [70.0, 70.0, 10.0, 10.0, 50.0, 10.0, 70.0, 35.0, 30.0, 10.0],
)
"""`(lower_bounds, upper_bounds)`"""
SUGGESTED_PARAMS = [
    67.84,
    57.57,
    5.98,
    5.68,
    2.66,
    -3.22,
    52.81,
    25.47,
    22.15,
    8.95,
]
"""Suggested parameters from the paper"""


def check_constrains(v: npt.NDArray[np.float32]) -> bool:
    """Return `False` if dimensions are invalid."""

    if v.shape[0] != N_DIMS_SLOTTED_PATCH:
        raise ValueError("v does not contain 10 elements.")
    w = v[0]
    L = v[1]
    wr = v[2]
    wu = v[3]
    lr = v[4]
    pr = v[5]
    lh = v[6]
    lv = v[7]
    fx = v[8]
    fy = v[9]
    if not fx < L / 2:
        return False
    if not fy < lv - wu:
        return False
    if not lh < L:
        return False
    if not lr < L:
        return False
    if not L < w / 2 - wr / 2 + pr:
        return False
    return True


def convert_to_variables(v: npt.NDArray[np.float32]) -> dict[str, str]:
    variables = {
        "h": "3.175 mm",
        "W": f"{v[0]} mm",
        "L": f"{v[1]} mm",
        "Wr": f"{v[2]} mm",
        "Wu": f"{v[3]} mm",
        "Lr": f"{v[4]} mm",
        "Pr": f"{v[5]} mm",
        "Lh": f"{v[6]} mm",
        "Lv": f"{v[7]} mm",
        "fx": f"{v[8]} mm",
        "fy": f"{v[9]} mm",
        "Lg": "L+6*h",
        "Wg": "W+6*h",
    }

    return variables


def create_slotted_patch(hfss: Hfss, variables: dict[str, str]) -> None:
    materials = ["pec", "Rogers RT/duroid 5880 (tm)", "Teflon (tm)"]

    hfss.solution_type = hfss.SolutionTypes.Hfss.DrivenModal
    hfss.set_auto_open()
    hfss.odesign.SetDesignSettings(  # pyright:ignore[reportOptionalMemberAccess]
        ["NAME:Design Settings Data", "Port Validation Settings:=", "Extended"]
    )
    # hfss.change_material_override()
    update_variables(hfss, variables)

    modeler = hfss.modeler
    assert isinstance(modeler, Modeler3D)

    check_materials(hfss, materials)

    current_objects = modeler.object_names  # pyright: ignore[reportUnknownVariableType]
    if "RadiatingSurface" in current_objects:
        current_objects.remove("RadiatingSurface")
    modeler.purge_history(current_objects)
    modeler.delete(current_objects)
    modeler.cleanup_objects()
    modeler.refresh()

    substrate = modeler.create_box(
        ["-Lg/2", "-Wg/2", "0 mm"],
        ["Lg", "Wg", "-h"],
        "substrate",
        "Rogers RT/duroid 5880 (tm)",
    )
    assert isinstance(substrate, Object3d)

    patch = modeler.create_rectangle(
        hfss.PLANE.XY, ["-L/2", "-W/2", "0 mm"], ["L", "W"], "patch"
    )
    assert isinstance(patch, Object3d)
    slot1 = modeler.create_rectangle(
        hfss.PLANE.XY, ["L/2", "Pr-Wr/2", "0 mm"], ["-Lr", "Wr"], "slot1"
    )
    slot2 = modeler.create_rectangle(
        hfss.PLANE.XY, ["-Lh/2", "-W/2+Lv-Wu", "0mm"], ["Lh", "Wu"], "slot2"
    )
    slot3 = modeler.create_rectangle(
        hfss.PLANE.XY, ["-Lh/2", "-W/2", "0 mm"], ["Wu", "Lv"], "slot3"
    )
    patch.subtract([slot1, slot2, slot3], False)
    hfss.assign_perfecte_to_sheets(patch, "patch")

    gnd = modeler.create_rectangle(
        hfss.PLANE.XY, ["-Lg/2", "-Wg/2", "-h"], ["Lg", "Wg"], "gnd"
    )
    assert isinstance(gnd, Object3d)
    hfss.assign_perfecte_to_sheets(gnd, "gnd")

    probe_in = modeler.create_cylinder(
        hfss.AXIS.Z,
        ["fx", "-W/2+fy", "0 mm"],
        "0.91 mm / 2",
        "-2*h",
        name="probe_in",
        matname="pec",
    )
    substrate.subtract(probe_in)
    probe_out = modeler.create_cylinder(
        hfss.AXIS.Z,
        ["fx", "-W/2+fy", "-h"],
        "3.58 mm / 2",
        "-h",
        name="probe_out",
        matname="pec",
    )
    assert isinstance(probe_out, Object3d)
    gnd.subtract(probe_out)
    probe_ins = modeler.create_cylinder(
        hfss.AXIS.Z,
        ["fx", "-W/2+fy", "-h"],
        "2.97 mm / 2",
        "-h",
        name="probe_ins",
        matname="Teflon (tm)",
    )
    assert isinstance(probe_ins, Object3d)
    probe_out.subtract(probe_ins)
    probe_ins.subtract(probe_in)
    port_face = probe_ins.bottom_face_z  # pyright: ignore[reportUnknownVariableType]
    assert isinstance(port_face, FacePrimitive)
    assert port_face.is_planar
    port = modeler.create_object_from_face(port_face)  # pyright: ignore[reportUnknownVariableType]
    hfss.create_lumped_port_to_sheet(
        port,
        [port_face.edges[0].midpoint, port_face.edges[1].midpoint],  # pyright: ignore
        portname="1",
        renorm=False,
    )
    # hfss.lumped_port(
    #     probe_in,
    #     probe_out,
    #     True,
    #     name="1",
    #     renormalize=False,
    # )

    setup_name = "Auto1"
    setup = hfss.get_setup(setup_name)
    assert isinstance(setup, SetupHFSS)
    setup.enable()

    setup.enable_adaptive_setup_multifrequency([1.9, 2.4], 0.02)
    setup.update({"MaximumPasses": 20})

    # sweep_name = "Sweep1"
    # sweep = setup.create_frequency_sweep(
    # "GHz", 1.5, 3, 401, sweep_name, sweep_type="Fast"
    # )
    # hfss.create_linear_count_sweep()

    # sweeps = h1.get_sweeps(setup_name)


def solve_sync(hfss: Hfss) -> SolutionData:
    """Synchronously solve and return solution data."""

    setup_name = "Auto1"
    setup = hfss.get_setup(setup_name)
    assert isinstance(setup, SetupHFSS)
    setup.analyze(10, 3, use_auto_settings=True)

    solution_data = setup.get_solution_data(  # pyright: ignore[reportUnknownVariableType]
        "dB(S(1,1))", "Sweep"
    )
    assert isinstance(solution_data, SolutionData)

    return solution_data


async def solve(hfss: Hfss) -> SolutionData:
    """Asynchronously solve and return solution data."""

    setup_name = "Auto1"
    setup = hfss.get_setup(setup_name)
    assert isinstance(setup, SetupHFSS)

    async def wait(t_start: float, t_mins: int):
        while hfss.are_there_simulations_running:
            await asyncio.sleep(5)
            t_current = time.time()
            t_mins_new = int((t_current - t_start) // 60)

            if t_mins_new > t_mins:
                t_mins = t_mins_new
                logger.info(
                    f"Simulation has been running for {t_mins} minutes."
                )

            if t_mins > 7:
                hfss.stop_simulations(False)
                logger.error(
                    "Simulation aborted for running for over 8 minutes."
                )
                while hfss.are_there_simulations_running:
                    await asyncio.sleep(5)
                    logger.debug("Waiting for simulation termination.")
                return

    while not setup.is_solved:
        t_start = time.time()
        t_mins = 0

        setup.analyze(10, 3, use_auto_settings=True, blocking=False)

        await wait(t_start, t_mins)

    solution_data = setup.get_solution_data(  # pyright: ignore[reportUnknownVariableType]
        "dB(S(1,1))", f"{setup_name} : LastAdaptive"
    )
    assert isinstance(solution_data, SolutionData)

    hfss.logger.clear_messages("", "")  # pyright: ignore[reportGeneralTypeIssues]

    return solution_data


def obj_fn_sync(hfss: Hfss, v: npt.NDArray[np.float32]) -> np.float32:
    """Object function (synchronous)."""

    variables = convert_to_variables(v)

    create_slotted_patch(hfss, variables)
    assert hfss.validate_full_design()[1]

    solution_data = solve_sync(hfss)

    s11 = solution_data.data_real()  # pyright: ignore[reportUnknownVariableType]
    assert isinstance(s11, list)

    hfss.cleanup_solution(entire_solution=False, field=True)

    return np.max(s11)


async def obj_fn(
    aedt_queue: asyncio.Queue[Hfss], v: npt.NDArray[np.float32]
) -> np.float32:
    """Object function (asynchronous)."""

    hfss = await aedt_queue.get()
    process_id = hfss.odesktop.GetProcessID()
    logger.debug(f"HFSS ({process_id}) receives task: {v}.")

    variables = convert_to_variables(v)

    create_slotted_patch(hfss, variables)
    assert hfss.validate_full_design()[1]

    t_start = time.time()

    solution_data = await solve(hfss)

    t_finish = time.time()

    time_str = time.strftime(
        r"%M min %S sec", time.localtime(t_finish - t_start)
    )

    s11 = solution_data.data_real()  # pyright: ignore[reportUnknownVariableType]
    assert isinstance(s11, list)

    hfss.cleanup_solution(entire_solution=False, field=True)

    logger.debug(f"HFSS ({process_id}) solved in {time_str}, {s11=}.")

    await aedt_queue.put(hfss)

    return np.max(s11)


# %%
if __name__ == "__main__":
    h1 = new_hfss_session()
    v = np.array(
        [67.84, 57.57, 5.98, 5.68, 2.66, -3.22, 52.81, 25.47, 22.15, 8.95]
    )
    print(obj_fn_sync(h1, v))



"""Utilities.
"""

import asyncio
from functools import wraps
from typing import Any, Callable, Coroutine

import numpy as np
import numpy.typing as npt
from loguru import logger
from pyaedt.hfss import Hfss

from antcal.pyaedt.hfss import new_hfss_session

TaskFn = Callable[
    [asyncio.Queue[Hfss], npt.NDArray[np.float32]],
    Coroutine[None, None, np.float32],
]
"""Task function signature"""


# %%
async def submit_tasks(
    task_fn: TaskFn,
    vs: npt.NDArray[np.float32],
    n_workers: int = 3,
    aedt_list: list[Hfss] | None = None,
) -> npt.NDArray[np.float32]:
    """Distribute simulation tasks to multiple AEDT sessions.

    :param task_fn: Task to run.
    :param vs: Input matrix, each row is one sample.
    :param n_workers: Number of AEDT to create, ignored if `aedt_queue` is provided.
    :param aedt_list: List of AEDT workers, for long running simulation tasks.
    :return: Results.
    """

    aedt_queue: asyncio.Queue[Hfss] = asyncio.Queue()

    if not aedt_list:
        logger.debug("aedt_list not provided, using self-hosted AEDT workers.")
        for _ in range(n_workers):
            await aedt_queue.put(new_hfss_session())
    else:
        logger.debug("Using provided aedt_list.")
        for hfss in aedt_list:
            await aedt_queue.put(hfss)

    async with asyncio.TaskGroup() as tg:
        tasks = [tg.create_task(task_fn(aedt_queue, v)) for v in vs]

    logger.debug("Simulation task queue completed.")
    results = np.array([task.result() for task in tasks])

    while not aedt_queue.empty():
        if not aedt_list:
            hfss = await aedt_queue.get()
            hfss.close_desktop()
        else:
            await aedt_queue.get()

    return results


async def refresh_aedt_queue(aedt_queue: asyncio.Queue[Hfss]) -> None:
    """Close all AEDTs in the queue and top it up with new ones."""

    n_simulators = aedt_queue.qsize()
    hfss = await aedt_queue.get()
    if hfss.desktop_class:
        non_graphical = hfss.desktop_class.non_graphical
    else:
        non_graphical = False
    hfss.close_desktop()

    while not aedt_queue.empty():
        hfss = await aedt_queue.get()
        hfss.close_desktop()

    for _ in range(n_simulators):
        await aedt_queue.put(new_hfss_session(non_graphical))


def refresh_aedt_list(aedt_list: list[Hfss]) -> None:
    """Close all AEDTs in the list and top it up with new ones."""

    n_simulators = len(aedt_list)
    if aedt_list[0].desktop_class:
        non_graphical = aedt_list[0].desktop_class.non_graphical
    else:
        non_graphical = False

    while len(aedt_list) > 0:
        hfss = aedt_list.pop()
        hfss.close_desktop()

    for _ in range(n_simulators):
        aedt_list.append(new_hfss_session(non_graphical))


# %%
def add_to_class(cls: type) -> Callable[..., Callable[..., Any]]:
    """A decorator that add the decorated function
    to a class as its attribute.

    In development, this decorator could be used to
    dynamically overwrite attributes in a class for
    convenience.

    The implementation came from [Michael Garod](https://mgarod.medium.com/dynamically-add-a-method-to-a-class-in-python-c49204b85bd6).

    :param cls: The class to be added to.

    :Examples:
    ```py
    class A:
        def __init__(self) -> None:
            ...

    @add_to_class(A)
    def print_hi(self: A) -> None:
        print("Hi")

    >>> a = A()
    >>> a.print_hi()
    Hi
    ```
    """

    def decorator(method: Callable[..., Any]) -> Callable[..., Any]:
        """This decorator perform the attachment,
        then just return the original function.
        """

        @wraps(method)
        def add_this(*args, **kwargs):  # pyright: ignore[reportMissingParameterType, reportUnknownParameterType]
            return method(*args, **kwargs)

        setattr(cls, method.__name__, add_this)
        return method

    return decorator



"""Helper functions around
{py:class}`pyaedt.hfss.Hfss`
for convenience.
"""

import sys
from types import MethodType
from typing import Mapping

from loguru import logger
from pyaedt.generic.desktop_sessions import (
    _desktop_sessions,  # pyright: ignore [reportPrivateUsage]
)
from pyaedt.generic.settings import settings
from pyaedt.hfss import Hfss


def __exit__(self: Hfss) -> None:
    """Release HFSS when leaving the context manager."""

    self.close_desktop()


def __del__(self: Hfss) -> None:
    """Release HFSS when there's no more reference."""

    self.close_desktop()


def close_desktop(self: Hfss) -> None:
    """Close desktop without saving the project."""

    try:
        self.close_project(save_project=False)
    except Exception as e:
        logger.error(f"Exception occurred during closing: {e}.")

    self.odesktop.QuitApplication()


def new_hfss_session(non_graphical: bool = False) -> Hfss:
    """Create a new HFSS instance, defaults to the latest version.

    A workaround to achieve multiple desktop sessions.

    :param bool non_graphical: Launch AEDT in non graphical mode, defaults to False
    :return Hfss: Hfss object

    :Examples:
    ```py
    >>> h1 = new_hfss_session()
    >>> h2 = new_hfss_session()
    ```
    """

    # Fallback to PythonNET
    settings.use_grpc_api = False
    # Reset desktop session tracker
    _desktop_sessions.clear()
    # Remove existing desktop handle
    try:
        del sys.modules["__main__"].oDesktop  # pyright: ignore
    except AttributeError:
        ...

    # Create a new HFSS object
    h = Hfss(non_graphical=non_graphical, new_desktop_session=True)

    # Rebind desktop properties
    d = sys.modules["__main__"].oDesktop
    desktop_install_dir = sys.modules["__main__"].sDesktopinstallDirectory
    h._odesktop = d  # pyright: ignore [reportPrivateUsage]
    # h._odesktop.aedt_version_id = h.odesktop.GetVersion()[0:6]  # pyright: ignore[reportPrivateUsage]
    h._desktop_install_dir = desktop_install_dir  # pyright: ignore [reportPrivateUsage]

    # Patch close methods
    h.close_desktop = MethodType(close_desktop, h)
    h.__exit__ = MethodType(__exit__, h)

    # My preferences
    # h.autosave_enable()
    h.autosave_disable()
    # h.logger.disable_stdout_log()  # pyright: ignore[reportGeneralTypeIssues]
    # h.change_material_override()

    return h


def get_variables(hfss: Hfss) -> dict[str, str]:
    vm = hfss.variable_manager
    if not vm:
        return {}
    return {k: v.evaluated_value for k, v in vm.design_variables.items()}


def update_variables(
    hfss: Hfss,
    variables: Mapping[str, str],
    constants: Mapping[str, str | float] | None = None,
) -> None:
    vm = hfss.variable_manager
    if not vm:
        return
    for item in variables.items():
        vm.set_variable(*item)
    if not constants:
        return
    for item in constants.items():
        vm.set_variable(*item)


def check_materials(hfss: Hfss, materials: str | list[str]) -> None:
    """If the material exists and is not in the materials database, it is added to this database."""

    mat = hfss.materials
    if isinstance(materials, str):
        materials = [materials]
    for material in materials:
        mat.checkifmaterialexists(material)



"""Decoupled DRAs: E-Plane.

Reference: C. Tong, N. Yang, K. W. Leung, P. Gu, and R. Chen,
"Port and Radiation Pattern Decoupling of Dielectric Resonator Antennas,"
_IEEE Transactions on Antennas and Propagation_,
vol. 70, no. 9, pp. 7713-7726, Sept. 2022,
doi: 10.1109/tap.2022.3177547.
"""

# %%
# import time

from ast import pattern
from importlib import reload
from typing import Literal, get_args

import numpy as np
import numpy.typing as npt
from loguru import logger
from pyaedt.hfss import Hfss
from pyaedt.modeler.cad.object3d import Object3d
from pyaedt.modeler.modeler3d import Modeler3D
from pyaedt.modules.Material import Material
from pyaedt.modules.SolveSetup import SetupHFSS
from pyaedt.modules.SolveSweeps import SweepHFSS
from pyaedt.modules.solutions import SolutionData

import antcal.pyaedt.hfss

reload(antcal.pyaedt.hfss)
from antcal.pyaedt.hfss import (
    check_materials,
    new_hfss_session,
    update_variables,
)

# %%
PLANES_LITERAL = Literal["e", "E", "h", "H"]
PARAMS_LIST = [
    "a",
    "b",
    "h",
    "l1",
    "w1",
    "l2",
    "w2",
    "l3",
    "w3",
    "ts",
    "ls",
    "ws",
    "lg",
    "wg",
    "l4",
    "w4",
]
SUGGESTED_PARAMS = np.array(
    [46, 23, 6.7, 5, 0.9, 2.5, 0.5, 1.8, 0.2, 1.524, 15, 2, 97, 50, 30, 3.6],
    dtype=np.float32,
)
CONSTANTS = {
    "epsilon_d": 9.9,
    "epsilon_s": 3.55,
    "tc": "0.035 mm",
    "freq_center": 4.9,
    "freq_start": 4.6,
    "freq_end": 5.2,
    "num_of_freq_points": 61,
    # "theta_start": -180,
    # "theta_stop": 180,
    # "theta_step": 1,
    # "phi_e_plane": 0,
    # "phi_h_plane": 90,
}
MATERIALS = ["pec", "copper"]


def convert_to_variables(v: npt.NDArray[np.float32]) -> dict[str, str]:
    return {p: f"{v:.4f} mm" for p, v in zip(PARAMS_LIST, v)}


def create_decoupled_dra_h(hfss: Hfss, variables: dict[str, str]) -> None:
    hfss.solution_type = hfss.SolutionTypes.Hfss.DrivenModal
    hfss.odesign.SetDesignSettings(  # pyright:ignore[reportOptionalMemberAccess]
        ["NAME:Design Settings Data", "Port Validation Settings:=", "Extended"]
    )
    hfss.set_auto_open(False)

    modeler = hfss.modeler
    assert isinstance(modeler, Modeler3D)

    current_objects = modeler.object_names
    if "RadiatingSurface" in current_objects:
        current_objects.remove("RadiatingSurface")
    modeler.purge_history(current_objects)
    modeler.delete(current_objects)
    modeler.refresh()

    hfss.set_auto_open()
    update_variables(hfss, variables, CONSTANTS)
    check_materials(hfss, MATERIALS)

    materials = hfss.materials
    mat_sub1 = materials.add_material("mat_sub1")
    assert isinstance(mat_sub1, Material)
    mat_sub1.permittivity = hfss["epsilon_s"]
    mat_sub1.dielectric_loss_tangent = 0.0027
    mat_sub2 = materials.add_material("mat_sub2")
    assert isinstance(mat_sub2, Material)
    mat_sub2.permittivity = hfss["epsilon_d"]

    sub1 = modeler.create_box(
        ["a/4-lg/2", "-wg/2", "-ts-tc"], ["lg", "wg", "ts"], "sub1", "mat_sub1"
    )
    assert isinstance(sub1, Object3d)

    gnd = modeler.create_object_from_face(sub1.top_face_z)
    assert isinstance(gnd, Object3d)

    slot1 = modeler.create_rectangle(
        hfss.PLANE.XY, ["-ls/2", "-ws/2", "-tc"], ["ls", "ws"], "slot1"
    )
    assert isinstance(slot1, Object3d)
    slot2 = slot1.duplicate_along_line(["a/2", 0, 0])
    assert isinstance(slot2, list)
    slot2.append(slot1)

    gnd.subtract(slot2, False)
    modeler.thicken_sheet(gnd, "tc")
    gnd.name = "gnd"
    gnd.material_name = "copper"

    modeler.create_box(["-a/4", "-b/2", 0], ["a", "b", "h"], "sub2", "mat_sub2")

    feed1 = modeler.create_box(
        ["-w4/2", "-wg/2", "-ts-tc*2"], ["w4", "l4", "tc"], "feed1", "copper"
    )
    assert isinstance(feed1, Object3d)
    feed2 = feed1.duplicate_along_line(["a/2", 0, 0])
    feed2 = modeler.get_object_from_name(feed2[0])
    assert isinstance(feed2, Object3d)

    port1 = modeler.create_rectangle(
        hfss.PLANE.ZX,
        ["-w4/2", "-wg/2", "-ts-tc*2"],
        ["ts+tc*2", "w4"],
        "port1",
    )
    assert isinstance(port1, Object3d)
    port2 = port1.duplicate_along_line(["a/2", 0, 0])
    port2 = modeler.get_object_from_name(port2[0])
    assert isinstance(port2, Object3d)
    hfss.create_lumped_port_to_sheet(
        port1, hfss.AxisDir.ZNeg, portname="1", renorm=False
    )
    hfss.create_lumped_port_to_sheet(
        port2, hfss.AxisDir.ZNeg, portname="2", renorm=False
    )

    strip1 = modeler.create_rectangle(
        hfss.PLANE.ZX, ["a/4-l2", "-b/2", 0], ["w2", "l1"], "strip1", "copper"
    )
    assert isinstance(strip1, Object3d)
    r2 = modeler.create_rectangle(
        hfss.PLANE.ZX, ["a/4-w3/2", "-b/2", "w2"], ["l3", "w3"], "copper"
    )
    r3 = modeler.create_rectangle(
        hfss.PLANE.ZX, ["a/4-l2", "-b/2", "w2+l3"], ["w1", "l1"], "copper"
    )
    strip1.unite([r2, r3])
    modeler.thicken_sheet(strip1, "-tc")
    strip1.duplicate_along_line([0, "b+tc"])

    modeler.cleanup_objects()

    setup_name = "Auto1"
    setup = hfss.get_setup(setup_name)
    assert isinstance(setup, SetupHFSS)
    setup.enable()
    setup.enable_adaptive_setup_single("4.9 GHz", 12, 0.05)
    setup.update(
        {
            "MinimumConvergedPasses": 2,
            "PercentRefinement": 45,
            "BasisOrder": -1,
            "DrivenSolverType": "Auto Select Direct/Iterative",
        }
    )

    sweep_name = "Sweep1"
    if sweep_name in hfss.get_sweeps(setup_name):
        setup.omodule.DeleteSweep(setup_name, sweep_name)
    sweep = setup.create_frequency_sweep(
        "GHz",
        CONSTANTS["freq_start"],  # type: ignore
        CONSTANTS["freq_end"],  # type: ignore
        CONSTANTS["num_of_freq_points"],
        sweep_name,
        sweep_type="Fast",
    )
    assert isinstance(sweep, SweepHFSS)

    logger.debug("Build completed.")


def solve_sync(hfss: Hfss) -> None:
    assert hfss.validate_full_design()[1]

    setup_name = "Auto1"

    setup = hfss.get_setup(setup_name)
    assert isinstance(setup, SetupHFSS)
    setup.analyze(10, 3, use_auto_settings=True)


def get_s_params(hfss: Hfss) -> SolutionData:
    setup_name = "Auto1"
    sweep_name = "Sweep1"

    setup = hfss.get_setup(setup_name)
    solution_data = setup.get_solution_data(
        domain="Sweep", sweep_name=sweep_name
    )
    assert isinstance(solution_data, SolutionData)

    return solution_data


def get_patterns(hfss: Hfss, plane: PLANES_LITERAL) -> SolutionData:
    if plane not in get_args(PLANES_LITERAL):
        raise TypeError(f'Plane "{plane}" is not one of {PLANES_LITERAL}')

    phi = "0deg" if plane.lower() == "e" else "90deg"

    setup_name = "Auto1"

    setup = hfss.get_setup(setup_name)
    solution_data = setup.get_solution_data(
        expressions=["GainPhi", "GainTheta"],
        report_category="Far Fields",
        variations={"Phi": phi},
        primary_sweep_variable="Theta",
        context="Elevation",
        sweep_name="LastAdaptive",
    )
    assert isinstance(solution_data, SolutionData)

    return solution_data


# %%
def run():
    variables = convert_to_variables(SUGGESTED_PARAMS)
    create_decoupled_dra_h(h1, variables)  # type: ignore  # noqa: F821

    logger.debug(h1.validate_full_design())

    solve_sync(h1)  # type: ignore # noqa: F821


if __name__ == "__main__":
    run()


# %%
def run2():
    patterns = get_patterns(h1, "h")

    import matplotlib.pyplot as plt

    plt.style.use(["default", "seaborn-v0_8-paper"])

    fig, ax = plt.subplots()
    ax.plot(patterns.primary_sweep_values, patterns.data_db10("GainPhi"))
    ax.plot(patterns.primary_sweep_values, patterns.data_db10("GainTheta"))
    fig.show()


if __name__ == "__main__":
    run2()


# %%
if __name__ == "__main__":
    h1 = new_hfss_session()
